// The MIT License

// Copyright (c) 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

import Connectable::*;
import GetPut::*;
import ClientServer::*;
import FIFO::*;
import SpecialFIFOs::*;

//AWB includes
`include "asim/provides/low_level_platform_interface.bsh"
`include "asim/provides/soft_connections.bsh"
`include "asim/provides/common_services.bsh"


// Local includes
`include "asim/provides/core.bsh"
`include "asim/provides/processor_library.bsh"
`include "asim/provides/processor_library.bsh"
`include "asim/provides/fpga_components.bsh"
`include "asim/rrr/remote_client_stub_PROCESSORSYSTEMRRR.bsh"

module [CONNECTED_MODULE] mkConnectedApplication ();

  Core core <- mkCore;
  Reg#(int) cycle <- mkReg(0);

  //External memory 
  // I'm not comfortable assuming that the memory subsystem is in order  
  // So I'll insert a completion buffer here.  
  ClientStub_PROCESSORSYSTEMRRR client_stub <- mkClientStub_PROCESSORSYSTEMRRR();   
  // Make this big enough so that several outstanding requests may be supported
  FIFO#(Bit#(MainMemTagSz)) tags <- mkSizedFIFO(8);

  // this is for the tracing
  rule printCycles;
    cycle <= cycle+1;
    $fdisplay(stderr, " => Cycle = %d", cycle);
  endrule


  rule sendMemReq;
    let coreReq <- core.mmem_client.request.get;
    case (coreReq) matches 
      tagged LoadReq .load: begin
                              //Allocate ROB space
                              client_stub.makeRequest_MemoryRequestLoad(load.addr);
                              tags.enq(load.tag);
                            end
      tagged StoreReq .store: begin
                                client_stub.makeRequest_MemoryRequestStore(store.addr,store.data);
                              end
    endcase
  endrule
  
  rule receiveMemResp;
    let memResp <- client_stub.getResponse_MemoryRequestLoad();
    tags.deq;
    core.mmem_client.response.put(tagged LoadResp {data:memResp,
                                                   tag: tags.first});
  endrule

endmodule